<?php

declare(strict_types = 1);

/**
 * Copyright (C) 2023 PRONOVIX GROUP.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

namespace Drupal\multi_env_vars_key_provider\Plugin\KeyProvider;

use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\key\KeyInterface;
use Drupal\key\Plugin\KeyPluginFormInterface;
use Drupal\key\Plugin\KeyPluginInterface;
use Drupal\key\Plugin\KeyPluginManager;
use Drupal\key\Plugin\KeyProviderBase;
use Drupal\key\Plugin\KeyTypeMultivalueInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Key provider that reads key values from multiple environment variables.
 *
 * @KeyProvider(
 *    id = "multi_env_vars_key_provider",
 *    label = @Translation("Multiple Environment Variables"),
 *    description = @Translation("Reads key values from multiple environment variables. Only supports multivalue key types."),
 *    tags = {
 *      "multivalue",
 *      "env",
 *    },
 *    key_value = {
 *      "accepted" = FALSE,
 *      "required" = FALSE
 *    }
 *  )
 *
 * @internal This class is not part of the module's public programming API.
 */
final class MultiEnvVarsKeyProvider extends KeyProviderBase implements KeyPluginFormInterface {

  /**
   * The key type manager.
   *
   * @var \Drupal\key\Plugin\KeyPluginManager
   *
   * @todo Mark this private after https://www.drupal.org/project/drupal/issues/3110266
   *   is merged and minimum required version is bumped to that Drupal core
   *   version.
   */
  protected KeyPluginManager $keyTypeManager;

  /**
   * The key input manager.
   *
   * @var \Drupal\key\Plugin\KeyPluginManager
   *
   * @todo Mark this private after https://www.drupal.org/project/drupal/issues/3110266
   *   is merged and minimum required version is bumped to that Drupal core
   *   version.
   */
  protected KeyPluginManager $keyInputManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): self {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);

    $instance->keyTypeManager = $container->get('plugin.manager.key.key_type');
    $instance->keyInputManager = $container->get('plugin.manager.key.key_input');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'env_var_prefix' => '',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $key_type = $this->keyTypeManager->createInstance($form_state->getUserInput()['key_type']);
    if (!$key_type instanceof KeyTypeMultivalueInterface) {
      $form_state->setErrorByName('key_type', $this->t('The selected key type is not a multivalue key type.'));
    }
    else {
      assert($key_type instanceof PluginInspectionInterface);
      $missing_env_vars = [];
      foreach ($key_type->getPluginDefinition()['multivalue']['fields'] as $name => $def) {
        // Fields assumed to be required unless it is specified otherwise.
        // @see \Drupal\key\Annotation\KeyType
        $required = TRUE;
        if (is_array($def)) {
          $required = $def['required'];
        }

        if ($required) {
          $env_var_value = $this->getEnvironmentVariableValueForField($form_state->getUserInput()['key_provider_settings']['env_var_prefix'], $name);
          if ($env_var_value === NULL) {
            $missing_env_vars[] = $this->generateEnvironmentVariableNameForField($form_state->getUserInput()['key_provider_settings']['env_var_prefix'], $name);
          }
        }
      }

      if ($missing_env_vars !== []) {
        $form_state->setErrorByName('key_provider', $this->formatPlural(
          count($missing_env_vars),
          'The %env_var environment variable is required but it is not set.',
          'The following environment variables are required but they are not set: %env_vars.',
          [
            '%env_var' => reset($missing_env_vars),
            '%env_vars' => implode(', ', $missing_env_vars),
          ]
        ));
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    // @todo Display an error message when the selected Key type is not a
    //   multi-value one. Hint: $form_state->triggering_element["#value"] or
    //   $form_state->input["key_type"] is the source for the selected key type.
    // @todo Improve UX by displaying the env var names on the UI.
    $form['env_var_prefix'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Environment variable prefix'),
      '#description' => $this->t("Prefix for all environment variables generated from the key type's field names."),
      '#default_value' => $this->getConfiguration()['env_var_prefix'],
      '#required' => TRUE,
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $this->setConfiguration($form_state->getValues());
  }

  /**
   * {@inheritdoc}
   */
  public function getKeyValue(KeyInterface $key): string {
    // Hint: This method is still called even if validateConfigurationForm()
    // failed.
    $key_type = $key->getKeyType();
    if ($key_type instanceof KeyTypeMultivalueInterface) {
      assert($key_type instanceof PluginInspectionInterface);

      // Inherit default field values from the related key input plugin.
      // An environment variable may not exist for non-required fields, but
      // we must return a field's default value if it was specified.
      $key_input_plugin = $this->keyInputManager->createInstance($key_type->getPluginDefinition()['key_value']['plugin']);
      assert($key_input_plugin instanceof KeyPluginInterface);
      $values = $key_input_plugin->getConfiguration();

      $env_var_prefix = $this->getConfiguration()['env_var_prefix'];
      foreach ($key_type->getPluginDefinition()['multivalue']['fields'] as $name => $def) {
        $env_var_value = $this->getEnvironmentVariableValueForField($env_var_prefix, $name);
        if ($env_var_value === NULL) {
          continue;
        }

        $values[$name] = $env_var_value;
      }

      return $key_type->serialize($values);
    }

    return '';
  }

  /**
   * Generates the environment variables name for a field.
   *
   * @param string $env_var_prefix
   *   The configured environment variable prefix.
   * @param string $field_name
   *   The field name.
   *
   * @return string
   *   The environment variable name.
   */
  private function generateEnvironmentVariableNameForField(string $env_var_prefix, string $field_name): string {
    return $env_var_prefix . '_' . strtoupper($field_name);
  }

  /**
   * Gets the value of a field from the related environment variable.
   *
   * @param string $env_var_prefix
   *   The configured environment variable prefix.
   * @param string $field_name
   *   The field name.
   *
   * @return string|null
   *   The environment variable value if set, otherwise null.
   */
  private function getEnvironmentVariableValueForField(string $env_var_prefix, string $field_name): ?string {
    $env_name = $this->generateEnvironmentVariableNameForField($env_var_prefix, $field_name);
    // @see https://github.com/symfony/symfony/blob/c8b962a0bf4ec24ef405222d66f3ae8e9eea94f6/src/Symfony/Component/DependencyInjection/EnvVarProcessor.php#L167
    $value = ($_ENV[$env_name] ?? $_SERVER[$env_name] ?? NULL) ?? getenv($env_name);
    return $value === FALSE ? NULL : $value;
  }

}
